import os
import sys
import numpy as np
from astropy.io import fits
from astropy import wcs
from astropy.wcs import WCS
from astropy.io.fits import update
from astropy.io.fits import getdata
import glob
from fitsio import FITS
from pathlib import Path

from astropy.table import QTable, Table, Column
from astropy import units as u

from astropy.coordinates import SkyCoord # High-level coordinates
from astropy.coordinates import ICRS, Galactic, FK4, FK5 # Low-level frames
from astropy.coordinates import Angle, Latitude, Longitude # Angles


import statistics
import shutil



def create_folder(folder):
    if not (os.path.exists(folder)):
        os.makedirs(folder)

def remove_file(filename):
    if(os.path.isfile(filename)==True):
        os.remove(filename)            

def do_evtool_esass(events=None,outfile=None,evlist=None,gti=None,region=None,emin=None,emax=None):

    eventfiles=None
    if(events):
        eventfiles="eventfiles=\'{}\'".format((" ").join(events))
    if(evlist):
        eventfiles="eventfiles=@{}".format(evlist)

    if not (eventfiles):
        print("ERROR: Event files not provided")

    emin="emin={}".format(emin) if(emin) else ''
    emax="emax={}".format(emax) if(emax) else ''
    gti="gti=\'{}\'".format(gti) if(gti) else ''
    region="region=\'{}\'".format(region) if(region) else ''

    cmd=["evtool",
         eventfiles,
         gti,region,emin,emax,
        "outfile={}".format(outfile),
        "image=yes",
        "flag=0x2000",
        "pattern=15"
        ]
    
    print((" ").join(cmd))
    test_exe('evtool')
    os.system((" ").join(cmd))

def set_bit(value, bit):
    return value | (1<<bit)

def clear_bit(value, bit):
    return value & ~(1<<bit)

def do_badpix_tm6(filename):
    """
    special corrrection for TM6 camera, needed for two CalPV-UDS observations,
    which shows high background at RAWX>250 
  
    creates new file with added _badpix.fits
    """

    nraw=384
    skip_after=250
    nskip=(nraw - skip_after)
    nrows0=nraw*nskip
    
    fout=filename.replace(".fits", "_badpix.fits")
    shutil.copyfile(filename, fout)

    hdul = fits.open(fout)
    #hdul.info()
    tstart=hdul[1].header['TSTART']
    tstop=hdul[1].header['TSTOP']
    nrows=hdul[1].header['NAXIS2']


    f = FITS(fout,'rw')
    data = f['EVENTS'].read()

    #print(data.dtype)

    for row in range(nrows):
        if (data['RAWX'][row] >= skip_after):
            data['FLAG'][row]=set_bit(data['FLAG'][row],13)
    #hdr = fits[0].read_header()
    f[1].write(data)
    f[1].write_comment("Changed by Roman Krivonos")
    f[1].write_history("All RAWx>=250 marked as BAD 0x2000")

    data_badpix = f['BADPIX6'].read()
    data4 = np.zeros(nrows0,dtype=data_badpix.dtype)

    n=0
    for i in range(skip_after,nraw):
        for j in range(nraw):
            data4['RAWX'][n]=i
            data4['RAWY'][n]=j
            data4['TIMEMIN'][n]=tstart
            data4['TIMEMAX'][n]=tstop
            data4['BADFLAG'][n]=3
            data4['TYPE'][n]=3
            data4['PHAMAX'][n]=12000
            data4['PHAMIN'][n]=0        
            n=n+1
        
    f[4].append(data4)
    f.close()

def init_events(key=None, eband_selected=[0], eband_index=None,
                ra_cen=None, de_cen=None,do_init=True,
                emin_kev=None, emax_kev=None, infile_dir=None, outfile_dir=None,
                do_obsmode=False,do_center=False,do_evtool=False,do_expmap=False):
    expmaps=[]
    expmap_all=[]
    infile_expmap=[]
    emin=[]
    emax=[]
    emin_ev_str=[]
    emax_ev_str=[]
    outfile_evtool=[]
    outfile_expmap=[]

    outfile_post='.fits'

    if not (infile_dir):
        print("infile_dir=?")
        return

    if not (outfile_dir):
        print("outfile_dir=?")
        return
    
    print("init events -- reading key {}".format(key))
    
    infile="{}/{}.fits".format(infile_dir,key)
        
    if(do_obsmode==True and do_init==True):
        """ correct OBS_MODE in files """
            
        lockfile="{}/{}.obsmode.lock".format(infile_dir,key)
        if not os.path.isfile(lockfile):
            with fits.open(infile) as hdu:
                for h in hdu:
                    extname = h.header['EXTNAME'] if "EXTNAME" in h.header else "None"
                    changeto = 'SURVEY' if ('scan' in infile) else 'POINTING'
                    if not (h.header['OBS_MODE'] == changeto):
                        print("*** {} {} --> {}".format(extname,h.header['OBS_MODE'],changeto))
                        h.header['OBS_MODE'] = changeto
                        hdu.writeto(infile,overwrite=True)
            Path(lockfile).touch()
        else:
            print("Lock file {} is found, skipping OBS_MODE correction.".format(lockfile))
            pass

    if(do_center==True and do_init==True):
        """ re-center original events files """
        if not (ra_cen and de_cen):
            print("Please provide center coordinates")
            sys.exit()
        cmd=["radec2xy", 
             "{}".format(infile), 
             "{}".format(ra_cen), 
             "{}".format(de_cen)]
            
        lockfile="{}/{}.radec2xy.lock".format(infile_dir,key)
        if not os.path.isfile(lockfile):
            print((" ").join(cmd))
            test_exe('radec2xy')
            os.system((" ").join(cmd))
            Path(lockfile).touch()
        else:
            print("Lock file {} is found, skipping radec2xy.".format(lockfile))
            pass


    outfile_evtool="{}/{}_EventList_en{}{}".format(outfile_dir, key, eband_index, outfile_post)

    cmd=["evtool",
         "eventfiles=%s" %(infile),
         "outfile=%s" %(outfile_evtool),
         "emin=%f" %(emin_kev),
         "emax=%f" %(emax_kev),
         #"region=%s" %(region),
         "image=yes",
         "flag=0x2000",
         "pattern=15"
        ]
    # run the command
    if(do_evtool==True and do_init==True):
        #log = subprocess.check_call(cmd)
        print((" ").join(cmd))
        test_exe('evtool')
        os.system((" ").join(cmd))
        """ correct OBS_MODE """
        with fits.open(outfile_evtool) as hdu:
            for h in hdu:
                extname = h.header['EXTNAME'] if "EXTNAME" in h.header else "None"
                changeto = 'SURVEY' if ('scan' in infile) else 'POINTING'
                if not (h.header['OBS_MODE'] == changeto):
                    print("*** {} {} --> {}".format(extname,h.header['OBS_MODE'],changeto))
                    h.header['OBS_MODE'] = changeto
                    hdu.writeto(outfile_evtool,overwrite=True)


    outfile_expmap="{}_ExposureMap_en{}{}".format(os.path.join(outfile_dir,key), eband_index, outfile_post)
    if(do_expmap==True and do_init==True):
        cmd=["expmap",
             "inputdatasets=%s" %(outfile_evtool),
             "emin=%s" %(emin_kev),
             "emax=%s" %(emax_kev),
             "templateimage=%s" %(outfile_evtool),  
             "singlemaps=%s" %(outfile_expmap),
             "withvignetting=yes",
             "withsinglemaps=yes","withfilebadpix=yes",
             "withmergedmaps=no",
             ]
        print((" ").join(cmd))
        test_exe('expmap')
        os.system((" ").join(cmd))

    return outfile_evtool,outfile_expmap


def test_exe(program):
    """ Tests if executable exists in PATH """
    import os
    def is_exe(fpath):
        return os.path.isfile(fpath) and os.access(fpath, os.X_OK)

    fpath, fname = os.path.split(program)
    if fpath:
        if is_exe(program):
            return program
    else:
        for path in os.environ["PATH"].split(os.pathsep):
            exe_file = os.path.join(path, program)
            if is_exe(exe_file):
                return exe_file

    print("\n*** Command {} not found ***\n".format(program))
    sys.exit()

def save_ds9reg(filename,scale=60):
    if(os.path.isfile(filename)==True):
        fout=filename.replace(".fits", ".reg")
        hdul = fits.open(filename)
        tbdata = hdul[1].data
        with open(fout, 'w') as writer:
            for rec in tbdata:
                writer.write("fk5;circle({}, {}, {})\n".format(rec['ra'],rec['dec'],rec['radec_err']/scale))

def save_ermldet_ds9reg(filename,scale=60,id_instr=1,id_band=1,ext_like=0.0,label='det_like'):
    if(os.path.isfile(filename)==True):
        #fout=filename.replace(".fits", ".instr{}.{}.reg".format(id_instr,label))
        fout=filename.replace(".fits", ".{}.reg".format(label))
        hdul = fits.open(filename)
        tbdata = hdul[1].data
        with open(fout, 'w') as writer:
            for rec in tbdata:
                if not (rec['id_inst'] == id_instr and rec['id_band'] == id_band):
                    continue
                if (rec['ext_like'] > ext_like):
                    continue
                if(abs(rec[label])>0.0):
                    writer.write("fk5;circle({}, {}, {}) # text={{{:.1f}}}\n".format(rec['ra'],rec['dec'],rec['radec_err']/scale,rec[label]))

def save_catprep_ds9reg(filename,scale=60,id_instr=1,id_band=1,ext_like=0.0,label='det_like_0'):
    if(os.path.isfile(filename)==True):
        #fout=filename.replace(".fits", ".instr{}.{}.reg".format(id_instr,label))
        fout=filename.replace(".fits", ".{}.reg".format(label))
        hdul = fits.open(filename)
        tbdata = hdul[1].data
        with open(fout, 'w') as writer:
            for rec in tbdata:
                if (rec['ext_like'] > ext_like):
                    continue
                if(abs(rec[label])>0.0):
                    writer.write("fk5;circle({}, {}, {}) # text={{{:.1f}}}\n".format(rec['ra'],rec['dec'],rec['radec_err']/scale,rec[label]))


def crossmatch_shu2019(filename,scale=1200,devmax=30,w=None,naxis1=1,naxis2=1,dlmin=6.0,dlmax=10000,ext_like=0.0,outkey='shu2019', catalog=None):
    if(os.path.isfile(filename)==False):
        print("File not found {}".format(filename))
    print("Start cross-match with Gaia-unWISE")

    if not catalog:
        print("ERROR: Please provide catalog.")
        return

    
    hdul = fits.open(catalog)
    tbdata_ref = hdul[1].data
    cat=[]
    for rec in tbdata_ref:
        cat.append({"ra":rec['ra'],"dec":rec['dec'],"gaia":rec["gaia_sourceid"],"unwise":rec["unwise_objid"]})


    hdul = fits.open(filename)
    tbdata_src = hdul[1].data

    srcs=[]
    for rec in tbdata_src:
        if (rec['ext_like'] > ext_like):
            continue
        srcs.append({"id":rec['id_src'],
                     "ra":rec['ra'],
                     "det_like":rec['det_like_0'],
                     "ext_like":rec['ext_like'],
                     "dec":rec['dec'],
                     "radec_err":rec['radec_err']/scale})

    fout=filename.replace(".fits", ".{}.cross".format(outkey))
    if(os.path.isfile(fout)==True):
        os.remove(fout)            

    t = Table(names=('ID','CNT',
                     'DET_LIKE', 'EXT_LIKE',  
                     'SRC_RA', 'SRC_DEC',
                     'REF_RA', 'REF_DEC',
                     'SEP','GAIA','UNWISE'), 
              dtype=('i4','i4','f4', 'f4','f8', 'f8','f8', 'f8','f4','i8','S16'))

    d = {}
    skip = {}

    # do correlation
    for src in srcs:
        src_crd = SkyCoord(src['ra'], src['dec'], frame=FK5(), unit="deg")
        sep=0
        
        for scat in cat:
            scat_crd = SkyCoord(scat['ra'], scat['dec'], frame=FK5(), unit="deg")
            sep = src_crd.separation(scat_crd).arcsec
            if(sep < devmax):
                sid=src['id']
                if(sid in d.keys()):
                    d[sid]=d[sid]+1
                    skip[sid]=1
                else:
                    d[sid]=1
                t.add_row((src['id'], d[sid], src['det_like'], src['ext_like'],
                           src['ra'], src['dec'],
                           scat['ra'], scat['dec'],
                           sep, scat['gaia'], scat['unwise']))
                
    t.write(fout, format='fits')

    src_tab = Table(names=('ID','RA', 'DEC'), dtype=('i4','f8','f8',))
    src_fout=fout.replace(".cross", ".{}.src".format(outkey))

    ref_fout=fout.replace(".cross", ".{}.ref".format(outkey))
    ref_tab = Table(names=('GAIA','RA', 'DEC'), dtype=('i8','f8','f8',))

    hdul = fits.open(fout)
    tbdata = hdul[1].data

    srcs=[]
    src_map=np.zeros((naxis1, naxis2))
    ref_map=np.zeros((naxis1, naxis2))
    delta_ra = []
    delta_dec = []
    for rec in tbdata:
        #if (rec['det_like'] < 10.0):
        #    continue
        if (rec['id'] in skip.keys()):
            print("Skip ID={} DET_LIKE={:.2f}".format(rec['id'],rec['det_like']))
        else:
            src_crd = SkyCoord(rec['src_ra'], rec['src_dec'], frame=FK5(), unit="deg")
            ref_crd = SkyCoord(rec['ref_ra'], rec['ref_dec'], frame=FK5(), unit="deg")
            src_x, src_y = np.round(w.world_to_pixel(src_crd))
            src_map[int(src_x), int(src_y)]=1
            ref_x, ref_y = np.round(w.world_to_pixel(ref_crd))
            ref_map[int(ref_x), int(ref_y)]=1
            src_tab.add_row((rec['id'],rec['src_ra'],rec['src_dec']))
            ref_tab.add_row((rec['gaia'],rec['ref_ra'],rec['ref_dec']))
            delta_ra.append(rec['src_ra']-rec['ref_ra'])
            delta_dec.append(rec['src_dec']-rec['ref_dec'])
            print("delta RA {:.6f} delta DEC {:.6f}".format(rec['src_ra']-rec['ref_ra'],
                            rec['src_dec']-rec['ref_dec']))

    print("Mean delta RA: {}, delta Dec: {} (arcsec)".format(statistics.mean(delta_ra)*3600,statistics.mean(delta_dec)*3600))
    src_tab.write(src_fout, format='fits', overwrite=True)
    with fits.open(src_fout) as f:
        f[0].verify('fix') 
        f[1].verify('fix') 
    with fits.open(src_fout, 'update') as f:
        f[0].header=f[0].header+w.to_header()
        f[0].data=src_map
        #f[1].header=f[1].header+w.to_header()

    ref_tab.write(ref_fout, format='fits', overwrite=True)
    with fits.open(ref_fout) as f:
        f[0].verify('fix') 
        f[1].verify('fix') 
    with fits.open(ref_fout, 'update') as f:
        f[0].header=f[0].header+w.to_header()
        f[0].data=ref_map
        #f[1].header=f[1].header+w.to_header()
